# Analyse Statique de Programmes -- TP Analyse Statique

CentraleSupélec

Enseignant: Virgile Prevosto

## Préliminaires

Ce TP est la suite directe du [précédent](tp.md), et vise à implémenter une
analyse des signes et une analyse des intervalles en plus de la propagation de
constante vue précédemment. En particulier, il est impératif d'avoir terminé
l'implémentation du calcul de point fixe (c'est-à-dire avoir complété la
fonction `fixpoint_iteration` du fichier [`iteration.py`](config/iteration.py))
pour pouvoir exécuter les exemples proposés dans ce TP.

### Rappel Docker
L'image est ici: https://github.com/Frederic-Boulanger-UPS/docker-webtop-3asl et peut être utilisée soit localement, soit depuis `MyDocker`. De plus le répertoire  `config` est monté automatiquement dans le container docker si vous utilisez les scripts associé ([PowerShell](https://github.com/Frederic-Boulanger-UPS/docker-webtop-3asl/blob/main/start-3asl.ps1) pour Windows ou [sh](https://github.com/Frederic-Boulanger-UPS/docker-webtop-3asl/blob/main/start-3asl.sh) pour Linux/macOS/BSD). Ces scripts devraient automatiquement ouvrir un onglet de votre navigateur web avec une session IceWM. Si ce n'est pas le cas, vous pouvez le faire manuellement: http://localhost:3000

NB: l'interpréteur Python est `python3` et non `python`.

### Typage statique
Les fichiers proposés ont des annotations de types permettant de vérifier que
les programmes sont bien typés avant de les exécuter. On pourra utiliser
l'outil [`pyright`](https://microsoft.github.io/pyright/#/) pour la vérification. Il n'est pas présent sur l'image
Docker, mais peut s'installer via les commandes suivantes:
```sh
sudo apt update
sudo apt install pipx
sudo apt install pyright
```

Il dispose également d'un plugin vscode.

## Treillis des signes

Dans un premier temps, on va s'intéresser à l'analyse des signes telle que vue en cours. Le but de cet exercice est de
compléter le fichier [`sign_analysis.py`](config/sign_analysis.py). Les différentes valeurs possibles sont définies dans
l'énumération `sign`, complétée par un élément `Top()` (une erreur a pu avoir lieu, mais n'est pas certaine), et `None` qui
indique comme précédemment l'absence de résultat, c'est à dire le fait qu'on a certainement une erreur. Une variante de `Top()`,
`BTop(True|False)` sera utilisée pour l'évaluation abstraite des expressions booléennes.

Comme pour la propagation de constante, les environnements abstraits sont des dictionnaires associant un `sign` à un nom de variable,
et un état est soit un environnement, soit `None` si on n'est jamais passé par le nœud correspondant.

On se donne également les opérations de treillis sur les `sign` (relation d'ordre, `join` et `meet`, ce dernier pouvant renvoyer
`None`), ainsi que les opérations arithmétiques abstraites.

### Évaluation des expressions

Compléter les fonctions `eval_aexp` et `eval_bexp`, et interpréter les résultats affichés par
[`example_sign_eval_exp.py`](config/example_sign_eval_exp.py).

NB: les opérations arithmétiques données travaillent uniquement sur des `sign`. C'est donc à `eval_aexp` de prendre en compte les
cas `None` et `Top`.

### Réduction d'un état

Compléter la fonction `reduce_state` qui étant donné un état et une expression booléenne essaie de réduire cet état sous l'hypothèse que
l'expression s'évalue à `True`. On pourra s'aider des fonctions définies au-dessus, qui prennent en argument un état, un nom de
variable `x`, et une expression arithmétique `expr`, et réduisent si possible la valeur associée à `x` en fonction de la comparaison
effectuée et de la valeur de `expr`. Le fichier [`example_sign_reduce.py`](config/example_sign_reduce.py) contient un certain nombre
de tests pour cette fonction.

### Fonctions de transfert

Compléter les fonctions de transfert de la classe `Sign_interp` et tester le résultat avec [`example_sign.py`](config/example_sign.py),
qui effectue l'analyse des signes pour le calcul de la factorielle

## Treillis des intervalles, début

On s'intéresse maintenant au treillis des intervalles. Le fichier [`interval_analysis.py`](config/interval_analysis.py) suit
la même structure que pour le treillis des signes, si ce n'est que les valeurs abstraites sont maintenant des objets de la classe
`Interval`, comportant deux champs, `lower_bound` et `upper_bound`, donc chacun peut être un entier ou `None`, ce qui indique qu'il
n'y a pas de borne.

On demande cette fois-ci de compléter les fonctions `interval_add` (addition d'intervalles) et `interval_div` (division d'intervalle).
Pour `interval_div`, on pourra notamment utiliser les fonctions `is_zero` et `contains_zero` qui renvoient `True` si leur argument est
respectivement le singleton `0` ou un intervalle contenant `0`. On pourra ensuite tester ces fonctions avec
[`example_interval_eval_exp.py`](config/example_interval_eval_exp.py)

Compléter ensuite les fonctions `reduce_eq`, `reduce_neq`, `reduce_leq` et `reduce_geq`, qui réduisent
la valeur d'une variable `x` sous l'hypothèse qu'elle est respectivement égale, différente, inférieure ou égale, ou supérieur ou égale à un entier donné. On pourra ensuite tester ces fonctions avec [`example_interval_reduce.py`](config/example_interval_reduce.py)

## Ajout de l'élargissement

Pour l'instant, notre fonction d'itération ne fait pas d'élargissement, donc l'analyse d'intervalle risque de ne pas terminer.
Modifier dans le fichier [`cfg.py`](config/cfg.py) la classe `LSkip` de manière à ce qu'elle ait un champ indiquant s'il s'agit
d'un arc de retour vers le nœud initial d'une boucle:

```python
@dataclass
class LSkip(Label):
    """label skip (does nothing)"""
    is_back_edge: bool
    def __str__(self):
        return "skip"
```

Dans ce même fichier, mettez à jour `cfg_aux` pour créer des labels `LSkip` avec le booléen adéquat

Modifier également le fichier [`iteration.py`](config/iteration.py) de la manière suivante:
- ajouter une fonction `widen` à la class `Transfer`, qui par défaut fait un simple `self.join`
- modifier `fixpoint_iteration` selon la version mise à jour de l'algorithme telle que montrée en cours.

On pourra tester l'algorithme avec le fichier [`example_widen.py`](config/example_widen.py), dont l'analyse compte le nombre maximum
d'arêtes rencontrées depuis le début du programme.

## Treillis des intervalles, version finale

Compléter la fonction `widen` et les fonctions de transfert de la classe `Interval_analysis`, et faire l'analyse d'intervalle du calcul de la factorielle avec le fichier [`example_interval.py`](config/example_interval.py)
